package no.ntnu.item.arctis.runtime;

import java.util.HashSet;
import java.util.Set;

/**
 * Java classes for building block partitions may extend this class to obtain
 * easier access to runtime features.
 * 
 * This class manages the ID of a block by implementing IArctisBuildingBlock.
 * 
 * @author kraemer
 * 
 */
public abstract class Block implements IArctisBuildingBlock {
	
	public static final String DEVICE_ID = "DEVICE_ID";

	/**
	 * The ID of this block. This value is set when a state machine is started,
	 * and must not to be written by application code.
	 */
	protected String blockID;
	protected String sessionID;

	/**
	 * Used by the runtime system to sets the ID of this block when a state
	 * machine is started, and must not to be used by application code.
	 */
	public void setBlockID(String blockID, String sessionID) {
		this.blockID = blockID;
		this.sessionID = sessionID;
	}

	/**
	 * Send an internal notification to this block with the given signal name.
	 * 
	 * Note that the suffix with the blockID for the signal name is optional, and
	 * will be appended if it is not present. If this is not desired, use the 
	 * method provided by AbstractRuntime.
	 * 
	 * @param signalName
	 */
	protected final void sendToBlock(String signalName) {
		if (signalName.endsWith(blockID) || signalName.contains(".")) {
			AbstractRuntime.getRuntime().sendToBlock(sessionID, blockID, signalName);
		} else {
			AbstractRuntime.getRuntime().sendToBlock(sessionID, blockID, signalName + "_" + blockID);
		}
	}

	/**
	 * Send an internal notification to this block with the given signal name.
	 * 
	 * Note that the suffix with the blockID for the signal name is optional, and
	 * will be appended if it is not present. If this is not desired, use the 
	 * method provided by AbstractRuntime.
	 * 
	 * @param signalName
	 * @param data
	 */
	protected final void sendToBlock(String signalName, Object data) {
		if (signalName.endsWith(blockID) || signalName.contains(".")) {
			AbstractRuntime.getRuntime().sendToBlock(sessionID, blockID, signalName, data);
		} else {
			AbstractRuntime.getRuntime().sendToBlock(sessionID, blockID, signalName + "_" + blockID, data);
		}
	}

	protected final void log(String message) {
		System.out.println(message);
	}

	/**
	 * Logs the given exception depending on the execution platform. 
	 * 
	 * @param message
	 * @param t
	 */
	protected final void log(String message, Throwable t) {
		// add information about this block in the future
		AbstractRuntime.getRuntime().catchApplicationException(message, t);
	}
	
	public static final void logException(String message, Throwable t) {
		AbstractRuntime.getRuntime().catchApplicationException(message, t);
	}
	
	public static interface LogListener {
		public void log(String subject, String message);
	}
	
	private static Set<LogListener> listeners = new HashSet<LogListener>();
	
	public static void notifyListeners(String subject, String message) {
		for(LogListener l: listeners) {
			l.log(subject, message);
		}
	}
	
	public static void addListener(LogListener l) {
		listeners.add(l);
	}
	
	public static void removeListener(LogListener l) {
		listeners.remove(l);
	}
	
	/**
	 * Operations within building block must never block. Long-running operations
	 * should be executed within an extra Runnable, executed via this method.
	 * 
	 * Use internal events to synchronize the execution of the long-running operation back
	 * to the logic of the building block.
	 * 
	 * This method creates a new thread. It is recommended to use this method instead of 
	 * creating own threads, since we may add logging and monitoring to this method in the future.
	 * 
	 * 
	 * @param r the Runnable containing the long-running operation to be executed
	 */
	public void runAsync(Runnable r) {
		new Thread(r).run();
	}
	
	public Object getProperty(String key) {
		if(!AbstractRuntime.context.containsKey(key)) {
			log("Property for key " + key + " does not exist.");
			return null;
		}
		return AbstractRuntime.context.get(key);
	}
	
	public boolean hasProperty(String key) {
		return AbstractRuntime.context.containsKey(key);
	}
	
	public static void setProperty(String key, Object value) {
		// TODO log setting properties
		AbstractRuntime.getRuntime().addContext(key, value);
	}

}
